/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *
 *  Copyright (C) 2005-2008 by TAKAGI Nobuhisa
 * 
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 */

/*!
 *  \file   toppers/text.hpp
 *  \brief  eLXgf[^ǗɊւ錾`
 *
 *  ̃t@CŒ`NX
 *  \code
 *  class basic_text< Container, CharT, Traits, Allocator >;
 *  \endcode
 */
#ifndef TOPPERS_TEXT_HPP_
#define TOPPERS_TEXT_HPP_

#include <algorithm>
#include <iterator>
#include <vector>
#include <numeric>
#include "toppers/text_line.hpp"
#include "toppers/misc.hpp"
#include <boost/utility.hpp>
#include <boost/iterator/iterator_facade.hpp>
#include <boost/format.hpp>

namespace toppers
{

  /*!
   *  \class  basic_text  text.hpp "toppers/text.hpp"
   *  \brief  eLXgf[^Ǘev[gNX
   *
   *  t@Cǂݍ񂾃eLXgf[^̊Ǘs߂̃NXłB
   *  ̃NXł́AeLXgf[^PȂ镶ƂĂł͂ȂA͌
   *  t@Cƍsԍt`ŊǗ܂B
   *
   *  \note   ̃NX̓eLXgҏW̗prz肵̂ł͂܂B
   */
  template
  <
    template < typename, class > class Container,
    typename CharT,
    class Traits = std::char_traits< CharT >,
    class Allocator = std::allocator< CharT >
  >
    class basic_text
  {
  public:
    typedef basic_text< Container, CharT, Traits, Allocator > self_t;
    typedef std::basic_string< CharT, Traits, Allocator > string_type;
    typedef ::toppers::text_line line_type;
    typedef basic_line_buf< CharT, Traits, Allocator > line_buf;

    //  ȉ͈ʓIȃReiƂ̌݊̂߂̌^`
    ////////////////////////////////////////////////////////////////////////////////
    typedef CharT value_type;
    typedef Container< line_buf, std::allocator< line_buf > > container;
    typedef value_type& reference;
    typedef value_type const& const_reference;
    typedef value_type* pointer;
    typedef value_type const* const_pointer;
    typedef Allocator allocator_type;
    typedef typename Allocator::difference_type difference_type;
    typedef typename Allocator::size_type size_type;

    /*!
     *  \class  const_iterator text.hpp "toppers/text.hpp"
     *  \brief  toppers::basic_text ̃Ce[^
     */
    class const_iterator : public boost::iterator_facade< const_iterator, CharT, boost::random_access_traversal_tag, CharT const& >
    {
      friend class boost::iterator_core_access;
    public:
      typedef boost::iterator_facade< const_iterator, CharT, boost::random_access_traversal_tag, CharT const& > base_t;
      typedef typename container::const_iterator row_iterator;
      typedef typename base_t::value_type value_type;
      typedef typename base_t::reference reference;
      typedef typename base_t::pointer pointer;
      typedef typename base_t::difference_type difference_type;

      const_iterator() : row_(), col_( 0 ) {}
      const_iterator( row_iterator row, size_type col = 0 ) : row_( row ), col_( col ) {}
      line_type const& line() const { return row_->line; }
      row_iterator const& get_row() const { return row_; }
      size_type get_col() const { return col_; }
    protected:
      reference dereference() const { return row_->buf[col_]; }
      bool equal( const_iterator const& rhs ) const { return row_ == rhs.row_ && col_ == rhs.col_; }
      void increment() { if ( ++col_ == row_->buf.size() ) { col_ = 0; ++row_; } }
      void decrement() { if ( col_ == 0 ) col_ = ( --row_ )->buf.size(); --col_; }
      void advance( difference_type n )
      {
        typename container::const_iterator row = row_;
        difference_type col = static_cast< difference_type >( col_ );
        if ( n >= 0 )
        {
          while ( static_cast< size_type>( col + n ) > row->buf.size() )
          {
            n -= row->buf.size() - col;
            ++row;
            col = 0;
          }
          col += n;
          if ( col == row->buf.size() )
          {
            col = 0;
            ++row;
          }
        }
        else  // n < 0
        {
          while ( col + n < -1 )
          {
            n += col;
            --row;
            col = row->buf.size() - 1;
          }
          col += n;
          if ( col == -1 )
          {
            --row;
            col = row->buf.size() - 1;
          }
        }
        row_ = row;
        col_ = col;
      }
      difference_type distance_to( const_iterator const& rhs ) const
      {
        difference_type n;
        typename container::const_iterator row = row_;
        if ( rhs.row_ < row_ )
        {
          n = -static_cast< difference_type >( col_ + ( rhs.row_->buf.size() - rhs.col_ ) );
          while ( --row != rhs.row_ )
          {
            n -= row->buf.size();
          }
        }
        else if ( row_ < rhs.row_ )
        {
          n = col_ + rhs.col_;
          while ( ++row != rhs.row_ )
          {
            n += row->buf.size();
          }
        }
        else
        {
          n = static_cast< difference_type >( rhs.col_ ) - static_cast< difference_type >( col_ );
        }
        return n;
      }
    private:
      typename container::const_iterator row_;
      size_type col_;
    };

    /*!
     *  \class  iterator text.hpp "toppers/text.hpp"
     *  \brief  toppers::basic_text ̃Ce[^
     */
    class iterator : public const_iterator
    {
    public:
      typedef typename container::iterator row_iterator;
      typedef typename const_iterator::value_type value_type;
      typedef typename const_iterator::reference reference;
      typedef typename const_iterator::pointer pointer;
      typedef typename const_iterator::difference_type difference_type;

      iterator() {}
      explicit iterator( row_iterator row, size_type col = 0 ) : const_iterator( row, col ) {}
      value_type& operator*() const { return const_cast< value_type& >( const_iterator::dereference() ); }
      iterator& operator++() { const_iterator::increment(); return *this; }
      iterator operator++( int ) { iterator t( *this ); const_iterator::increment(); return t; }
      iterator& operator--() { const_iterator::decrement(); return *this; }
      iterator operator--( int ) { iterator t( *this ); const_iterator::decrement(); return t; }
    };

    typedef std::reverse_iterator< const_iterator > const_reverse_iterator;
    typedef std::reverse_iterator< iterator > reverse_iterator;

  public:
    /*!
     *  \brief  ftHgRXgN^
     */
    basic_text() {}

    /*!
     *  \brief  RXgN^
     *  \param  first   ɗpf[^̐擪
     *  \param  last    ɗpf[^̏I[+1
     */
    basic_text( const_iterator const& first, const_iterator const& last )
    {
      if ( first != last )
      {
        init( first, last );
      }
    }

    /*!
     *  \brief  RXgN^
     *  \param  first   ɗpf[^̐擪
     *  \param  last    ɗpf[^̏I[+1
     *  \note   t@C"unknown"ɂȂ܂B
     */
    template < class InputIterator >
      basic_text( InputIterator first, InputIterator last )
    {
      if ( first != last )
      {
        init( first, last, append_directive );
      }
    }

    /*!
     *  \brief  RXgN^
     *  \param  first   ɗpf[^̐擪
     *  \param  last    ɗpf[^̏I[+1
     *  \param  directive PsReiɊi[邽߂̏
     *  \note   t@C"unknown"ɂȂ܂B
     *
     *  directive  void directive( container& cont, line_buf& buf )̌`
     *  Ƃ֐ւ̃|C^܂̓t@N^ł邱Ƃz肵Ă܂B
     *  cont ͓Ńf[^̊i[ɎgpReiA buf ͂Ps̕
     *  ƃt@CEsԍi[obt@łB
     */
    template < class InputIterator, class Directive >
      basic_text( InputIterator first, InputIterator last, Directive directive )
    {
      if ( first != last )
      {
        init( first, last, directive );
      }
    }

    /*!
     *  \brief  RXgN^
     *  \param  istr    ɗp̓Xg[
     *  \note   t@C"unknown"ɂȂ܂B
     */
    basic_text( std::basic_istream< CharT, Traits >& istr )
    {
      init( istr, append_directive );
    }

    /*!
     *  \brief  RXgN^
     *  \param  istr    ɗp̓Xg[
     *  \param  directive PsReiɊi[邽߂̏
     *  \note   t@C"unknown"ɂȂ܂B
     *
     *  directive  void directive( container& cont, line_buf& buf )̌`
     *  Ƃ֐ւ̃|C^܂̓t@N^ł邱Ƃz肵Ă܂B
     *  cont ͓Ńf[^̊i[ɎgpReiA buf ͂Ps̕
     *  ƃt@CEsԍi[obt@łB
     */
    template < class Directive >
      basic_text( std::basic_istream< CharT, Traits >& istr, Directive directive )
    {
      init( istr, directive );
    }

    /*!
     *  \brief  擪ʒu̎擾
     *  \return 擪ւ̃Ce[^Ԃ
     */
    const_iterator begin() const
    {
      return const_iterator( container_.begin() );
    }

    /*!
     *  \brief  擪ʒu̎擾
     *  \return 擪ւ̃Ce[^Ԃ
     */
    iterator begin()
    {
      return iterator( container_.begin() );
    }
    
    /*!
     *  \brief  s̎擾
     *  \param  pos  擪̃ItZbg
     *  \return posŎw肳ꂽItZbgs߁CCe[^Ԃ
     */
    const_iterator line_at(size_type pos) const
    {
        typedef text::container::const_iterator const_row_iterator;
        const_iterator first( container_.begin() ), last( container_.end() );
        int count = 0 , buf_size , i;
        size_type sum = 0;

        while((sum < pos) && (first != last))
        {
            const_row_iterator current( first.get_row() );
            buf_size = current->buf.size();
            sum += buf_size;
            i = 0;
            while(i < buf_size)
            {
                i++;
                first++;
            }
            count += 1;
        }
        
        if(first < last)
        {
            first -= buf_size;
            return const_iterator(first);
        }
        else
        {
            return const_iterator( --last );
        }
    }   
    /*!
     *  \brief  I[ʒu+1̎擾
     *  \return I[̎vfւ̃Ce[^Ԃ
     */
    const_iterator end() const
    {
      return const_iterator( container_.end() );
    }

    /*!
     *  \brief  I[ʒu+1̎擾
     *  \return I[̎vfւ̃Ce[^Ԃ
     */
    iterator end()
    {
      return iterator( container_.end() );
    }

    /*!
     *  \brief  t擪ʒu̎擾
     *  \return t擪ւ̋tCe[^Ԃ
     */
    const_reverse_iterator rbegin() const
    {
      string_type const* pbuf = &container_.back().buf;
      return const_reverse_iterator( container_.rbegin() + pbuf->empty() ? 0 : pbuf->size()-1 );
    }

    /*!
     *  \brief  t擪ʒu̎擾
     *  \return t擪ւ̋tCe[^Ԃ
     */
    reverse_iterator rbegin()
    {
      string_type* pbuf = &container_.back().buf;
      return reverse_iterator( container_.rbegin() + pbuf->empty() ? 0 : pbuf->size()-1 );
    }

    /*!
     *  \brief  tI[ʒu+1̎擾
     *  \return tI[̎vfւ̋tCe[^Ԃ
     */
    const_reverse_iterator rend() const
    {
      return const_reverse_iterator( container_.rend() );
    }

    /*!
     *  \brief  tI[ʒu+1̎擾
     *  \return tI[̎vfւ̋tCe[^Ԃ
     */
    reverse_iterator rend()
    {
      return reverse_iterator( container_.rend() );
    }

    /*!
     *  \brief  擪̎Q
     *  \return 擪ւ̎QƂԂ
     */
    const_reference front() const
    {
      return container_.front().buf[0];
    }

    /*!
     *  \brief  擪̎Q
     *  \return 擪ւ̎QƂԂ
     */
    reference front()
    {
      return container_.front().buf[0];
    }

    /*!
     *  \brief  I[̎Q
     *  \return I[ւ̎QƂԂ
     */
    const_reference back() const
    {
      string_type const* pbuf = &container_.back().buf;
      return pbuf->at( pbuf->size()-1 );
    }

    /*!
     *  \brief  I[̎Q
     *  \return I[ւ̎QƂԂ
     */
    reference back()
    {
      string_type* pbuf = &container_.back().buf;
      return pbuf->at( pbuf->size()-1 );
    }

    /*!
     *  \brief  wʒu̎̕Q
     *  \return wʒu̕ւ̎QƂԂ
     */
    const_reference at( size_type pos ) const
    {
      return container_.front().buf[pos];
    }

    /*!
     *  \brief  wʒu̎̕Q
     *  \return wʒu̕ւ̎QƂԂ
     */
    reference at( size_type pos )
    {
      return container_.front().buf[pos];
    }

    /*!
     *  \brief  w͈͂̃eLXgf[^ǉ
     *  \param  first ǉɗpf[^̐擪
     *  \param  last  ǉɗpf[^̏I[+1
     */
    void append( const_iterator const& first, const_iterator const& last )
    {
      if ( first != last )
      {
        self_t t( *this );
        t.init( first, last );
        swap( t );
      }
    }

    /*!
     *  \brief  w͈͂̃eLXgf[^ǉ
     *  \param  first ǉɗpf[^̐擪
     *  \param  last  ǉɗpf[^̏I[+1
     *  \note   t@Cƍsԍ͒ǉ̏I[̂̂p܂B
     *          ǉɃf[^Ȃꍇ"unknown"ɂȂ܂B
     */
    template < class InputIterator >
      void append( InputIterator first, InputIterator last )
    {
      append( first, last, append_directive );
    }

    /*!
     *  \brief  w͈͂̃eLXgf[^ǉ
     *  \param  first ǉɗpf[^̐擪
     *  \param  last  ǉɗpf[^̏I[+1
     *  \param  directive PsReiɊi[邽߂̏
     *  \note   t@Cƍsԍ͒ǉ̏I[̂̂p܂B
     *          ǉɃf[^Ȃꍇ"unknown"ɂȂ܂B
     *
     *  directive  void directive( container& cont, line_buf& buf )̌`
     *  Ƃ֐ւ̃|C^܂̓t@N^ł邱Ƃz肵Ă܂B
     *  cont ͓Ńf[^̊i[ɎgpReiA buf ͂Ps̕
     *  ƃt@CEsԍi[obt@łB
     */
    template < class InputIterator, class Directive >
      void append( InputIterator first, InputIterator last, Directive directive )
    {
      if ( first != last )
      {
        self_t t( *this );
        t.init( first, last, directive );
        swap( t );
      }
    }

    /*!
     *  \brief  w͈͂̃eLXgf[^ǉ
     *  \param  istr      ǉɗp̓Xg[
     *  \note   t@Cƍsԍ͒ǉ̏I[̂̂p܂B
     *          ǉɃf[^Ȃꍇ"unknown"ɂȂ܂B
     */
    void append( std::basic_istream< CharT, Traits >& istr )
    {
      append( istr, append_directive );
    }

    /*!
     *  \brief  w͈͂̃eLXgf[^ǉ
     *  \param  istr      ǉɗp̓Xg[
     *  \param  directive PsReiɊi[邽߂̏
     *  \note   t@Cƍsԍ͒ǉ̏I[̂̂p܂B
     *          ǉɃf[^Ȃꍇ"unknown"ɂȂ܂B
     *
     *  directive  void directive( container& cont, line_buf& buf )̌`
     *  Ƃ֐ւ̃|C^܂̓t@N^ł邱Ƃz肵Ă܂B
     *  cont ͓Ńf[^̊i[ɎgpReiA buf ͂Ps̕
     *  ƃt@CEsԍi[obt@łB
     */
    template < class Directive >
      void append( std::basic_istream< CharT, Traits >& istr, Directive directive )
    {
      self_t t( *this );
      t.init( istr, directive );
      swap( t );
    }

    /*!
     *  \brief  w͈͂̃eLXgf[^
     *  \param  first ɗpf[^̐擪
     *  \param  last  ɗpf[^̏I[+1
     */
    void assign( const_iterator const& first, const_iterator const& last )
    {
      self_t t( first, last );
      swap( t );
    }

    /*!
     *  \brief  w͈͂̃eLXgf[^
     *  \param  first ɗpf[^̐擪
     *  \param  last  ɗpf[^̏I[+1
     *  \note   t@C"unknown"ɂȂ܂B
     */
    template < class InputIterator >
      void assign( InputIterator first, InputIterator last )
    {
      self_t t( first, last, append_directive );
      swap( t );
    }

    /*!
     *  \brief  w͈͂̃eLXgf[^
     *  \param  first ɗpf[^̐擪
     *  \param  last  ɗpf[^̏I[+1
     *  \param  directive PsReiɊi[邽߂̏
     *  \note   t@C"unknown"ɂȂ܂B
     *
     *  directive  void directive( container& cont, line_buf& buf )̌`
     *  Ƃ֐ւ̃|C^܂̓t@N^ł邱Ƃz肵Ă܂B
     *  cont ͓Ńf[^̊i[ɎgpReiA buf ͂Ps̕
     *  ƃt@CEsԍi[obt@łB
     */
    template < class InputIterator, class Directive >
      void assign( InputIterator first, InputIterator last, Directive directive )
    {
      self_t t( first, last, directive );
      swap( t );
    }

    /*!
     *  \brief  w͈͂̃eLXgf[^
     *  \param  istr      ɗp̓Xg[
     *  \note   t@Cƍsԍ͑̏I[̂̂p܂B
     *          ǉɃf[^Ȃꍇ"unknown"ɂȂ܂B
     */
    void assign( std::basic_istream< CharT, Traits >& istr )
    {
      self_t t( istr, append_directive );
      swap( t );
    }

    /*!
     *  \brief  w͈͂̃eLXgf[^
     *  \param  istr      ɗp̓Xg[
     *  \param  directive PsReiɊi[邽߂̏
     *  \note   t@Cƍsԍ͑̏I[̂̂p܂B
     *          ǉɃf[^Ȃꍇ"unknown"ɂȂ܂B
     *
     *  directive  void directive( container& cont, line_buf& buf )̌`
     *  Ƃ֐ւ̃|C^܂̓t@N^ł邱Ƃz肵Ă܂B
     *  cont ͓Ńf[^̊i[ɎgpReiA buf ͂Ps̕
     *  ƃt@CEsԍi[obt@łB
     */
    template < class Directive >
      void assign( std::basic_istream< CharT, Traits >& istr, Directive directive )
    {
      self_t t( istr, directive );
      swap( t );
    }

    /*!
     *  \brief  I[ɕǉ
     *  \param  value ǉ镶
     */
    void push_back( value_type value )
    {
      if ( container_.empty() || container_.back().line.file != line_.file )
      {
        line_buf buf( initial_line(), string_type( 1, value ) );
        container_.push_back( buf );
      }
      else
      {
        string_type const* pbuf = &container_.back().buf;
        if ( pbuf->at( pbuf->size()-1 ) == widen( '\n' ) )
        {
          ++line_.line;
          line_buf buf( initial_line(), string_type( 1, value ) );
          container_.push_back( buf );
        }
        else
        {
          container_.back().buf += value;
        }
      }
    }

    /*!
     *  \brief  I[當폜
     */
    void pop_back()
    {
      if ( container_.empty() )
      {
        container_.pop_back();
      }
      else
      {
        string_type const* pbuf = &container_.back().buf;
        pbuf->resize( pbuf->size()-1 );
        if ( container_.back().buf.empty() )
        {
          container_.pop_back();
        }
      }
    }

    /*!
     *  \brief  Rei󂩂ǂ̔
     *  \return ̏ꍇ true Ԃ
     */
    bool empty() const
    {
      return container_.empty();
    }

    /*!
     *  \brief  ReĩTCY
     *  \return ReiɊ܂܂Ă镶Ԃ
     */
    size_type size() const
    {
      return std::accumulate( container_.begin(), container_.end(), size_type( 0 ), add_line_size );
    }

    /*!
     *  \brief  Reǐ
     *  \param  other   ΏۂƂȂRei
     */
    void swap( self_t& other )
    {
      container_.swap( other.container_ );
      line_.swap( other.line_ );
    }

    /*!
     *  \brief  ReĩNA
     */
    void clear()
    {
      container_.clear();
      line_.file.clear();
      line_.line = 0;
    }

    void set_line( std::string const& file, long line )
    {
      line_.file = file;
      line_.line = line;
    }

    line_type get_line() const
    {
      return line_;
    }
  private:
    //! ̃Rei͌Ƃ鏉
    void init( const_iterator const& first, const_iterator const& last )
    {
      line_buf buf( *first.get_row() );
      if ( first.get_col() > 0 )
      {
        buf.buf = buf.buf.substr( first.get_col() );
      }
      container_.push_back( buf );
      std::copy( boost::next( first ).get_row(), last.get_row(), std::back_inserter( container_ ) );
      buf = *last.get_row();
      buf.buf = buf.buf.substr( 0, last.get_col() );
      container_.push_back( buf );
      line_ = buf.line;
    }

    //! َ̃Rei͌Ƃ鏉
    template < class InputIterator, class Directive >
      void init( InputIterator first, InputIterator last, Directive directive )
    {
      line_buf buf( initial_line() );
      value_type const nl = widen( '\n' );
      for ( InputIterator iter = first; iter != last; ++iter )
      {
        value_type ch = *iter;
        buf.buf += ch;
        if ( ch == nl )
        {
          directive( container_, buf );
          line_ = buf.line;
        }
      }
      if ( !buf.buf.empty() )
      {
        container_.push_back( buf );
      }
      line_ = buf.line;
    }

    //! ̓Xg[͌Ƃ鏉
    template < class Directive >
      void init( std::basic_istream< CharT, Traits >& istr, Directive directive )
    {
      line_buf buf( initial_line() );
      value_type const nl = widen( '\n' );
      while ( istr )
      {
        value_type ch;
        istr.get( ch );
        buf.buf += ch;
        if ( ch == nl )
        {
          directive( container_, buf );
          line_ = buf.line;
        }
      }
      if ( !buf.buf.empty() )
      {
        container_.push_back( buf );
      }
      line_ = buf.line;
    }

    //! t@CEsԍ̏̎擾
    line_type const initial_line() const
    {
      if ( line_.line > 0 )
      {
        return line_;
      }
      if ( container_.empty() )
      {
        return line_type( "unknown", 1 );
      }
      return container_.back().line;
    }

    //! VOoCgʌ`̕^ւ̕ϊ
    static value_type widen( char ch )
    {
      return ::toppers::widen< value_type >( ch );
    }

    //! size ֐ŎgpݎZ
    static size_type add_line_size( size_type value, line_buf const& buf )
    {
      return value + buf.buf.size();
    }

    //! Directive ̃ftHg
    static void append_directive( container& cont, line_buf& buf )
    {
      cont.push_back( buf );
      ++buf.line.line;
      buf.buf.clear();
    }
  private:
    container container_;
    line_type line_;
  };

  typedef basic_text< std::vector, char > text;

  inline text_line const& get_text_line( text::const_iterator iter )
  {
    return iter.line();
  }

  inline text_line const& get_text_line( text::iterator iter )
  {
    return iter.line();
  }

}

#endif  // ! TOPPERS_TEXT_HPP_
